#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

import Tkinter as tk
import threading
import Queue
import sys
import gc
import time

#!
#
# スティック操作量表示用GUIのクラス
#
# start() / pause() の呼び出しにより OpenRTM の RTC の activation / deactivation に
# 連動して ウィンドウの表示/非表示を切り替える。
#
# このクラスはTk処理のスレッドを停止させずにウィンドウの表示/非表示のみを切り換える
# ように動作する。この設計が必要となるのは，Tkinter の仕様上，最初にTkinterを初期化した
# スレッドと異なるスレッドから新たにトップレベルウィンドウを表示しようとすると処理が
# ブロックしてしまうためである。
#
class JoyVelCtrlWidget():
    def __init__(self, helpText, minFwdVel, maxFwdVel, minAngVel, maxAngVel):
        # 以下の数値が大きすぎるとジョイスティック等で外部から setScaleValues() 
        # で値を設定したときのスケールの追従性が不十分
        self.__UPDATE_PERIOD = 50 # [msec]
        
        self.__helpText  = helpText
        self.__minFwdVel = minFwdVel
        self.__maxFwdVel = maxFwdVel
        self.__minAngVel = minAngVel
        self.__maxAngVel = maxAngVel
        self.__refVelQueue = Queue.Queue()
        self.__pauseRequested   = threading.Event()
        self.__destroyRequested = threading.Event()
        self.__iconifyCompleted = threading.Condition()
        self.__tkMainloopReady  = threading.Condition()

        self.__startRequested = threading.Condition()

        self.__pauseRequested.set() # ウィンドウを表示しない状態を開始状態とする
        self.__widgetsCreated = False

         # Tk に直接アクセスするスレッドを一つにするため別のスレッドを準備
        print "starting a thread for Tk..."
        self.__tkloopThread = threading.Thread(target=self.threadProc)
        self.__tkloopThread.start()

    def __del__(self):
        self.destroy()

    def createWidgets(self, helpText, minFwdVel, maxFwdVel, minAngVel, maxAngVel):
        SCALE_RES = 0.01
        
        print "calling tk.Tk()..."
        self.__root = tk.Tk()

        # Tk処理スレッドで問題が起きないように，閉じるボタンに対する
        # デフォルトのコールバックは無効にする
        self.__root.protocol("WM_DELETE_WINDOW", self.deleteCallback)

        self.__root.title('JoystickVelocityControl')
        self.__root.grid()
        
        self.__scales = dict()
        self.__scales['fwd vel'] = tk.Scale(self.__root, orient=tk.VERTICAL,
                                          from_=maxFwdVel, to=minFwdVel,
                                          resolution=SCALE_RES)
        
        self.__scales['ang vel'] = tk.Scale(self.__root, orient=tk.VERTICAL,
                                          from_=maxAngVel, to=minAngVel,
                                          resolution=SCALE_RES) 

        numRows = 0
        
        self.__scales['fwd vel'].grid(row=numRows, column=0)
        self.__scales['ang vel'].grid(row=numRows, column=1)
        numRows += 1
        
        tk.Label(self.__root, width=15, text='fwd vel [m/s]').grid(row=numRows, column=0)
        tk.Label(self.__root, width=15, text='ang vel [rad/s]').grid(row=numRows, column=1)
        numRows += 1
        
        #sc.grid()

        lb = tk.Label(self.__root,
                      text=helpText,
                      relief='ridge', justify=tk.LEFT).grid(row=numRows, columnspan=4)

    def __del__(self):
        del self.__root
        del self.__tkloopThread

    def start(self):
        self.__pauseRequested.clear()
        
        self.__startRequested.acquire()
        self.__startRequested.notify()
        self.__startRequested.release()

        # 待たないとウィンドウの表示処理が確定的にならないので待つ
        self.__tkMainloopReady.acquire()
        self.__tkMainloopReady.wait() 
        self.__tkMainloopReady.release()
        

    def threadProc(self):
        while True: # not self.__destroyRequested.is_set():
            self.__startRequested.acquire()
            self.__startRequested.wait()
            self.__startRequested.release()

            if self.__destroyRequested.is_set():
                break
            
            print "starting Tk processing..."
            if not self.__widgetsCreated:
                print "calling createWidgets()..."
                self.createWidgets(self.__helpText,
                                   self.__minFwdVel, self.__maxFwdVel,
                                   self.__minAngVel, self.__maxAngVel)
                self.__widgetsCreated = True
                print "widgets creation complete"
            
            self.__root.after(self.__UPDATE_PERIOD, self.updateWidgets)
            
            self.__root.deiconify()
            self.__root.lift()

            self.__tkMainloopReady.acquire()
            self.__tkMainloopReady.notify() 
            self.__tkMainloopReady.release()

            print "mainloop()..."
            self.__root.mainloop()

            # 制御が以下に到達するのは self.__pauseRequested が set されたあと
            print "iconifying..."
            self.__root.iconify()
            # Tk の仕様上 invisible に設定できないので代わりに iconify

            self.__iconifyCompleted.acquire()
            self.__iconifyCompleted.notify()
            self.__iconifyCompleted.release()
                
            #if not self.__pauseRequested.is_set():
            #   do tk processing...    
            # else:
            #     print "Tk thread sleeping..."
            #     SLEEP_PERIOD = 1 # [sec]
            #     time.sleep(SLEEP_PERIOD)

    # mainloop() のスレッドからコールバックされる
    def updateWidgets(self):
        # print "updateWidgets()"
        try:
            fwdVel, angVel = self.__refVelQueue.get(block=False)
        except Queue.Empty:
            pass
        else:
            self.__scales['fwd vel'].set(fwdVel)
            self.__scales['ang vel'].set(angVel)

        if self.__pauseRequested.is_set():
            #self.__root.after_cancel(self.__updateCallbackID)
            self.__root.quit()
        else:
            self.__root.after(self.__UPDATE_PERIOD, self.updateWidgets)
        
    def deleteCallback(self):
        print "deleteCallback()"
        # self.pause()

    def setScaleValues(self, fwdVel, angVel):
        # print thr, rol, pit, yaw
        self.__refVelQueue.put((fwdVel, angVel))
        
    def pause(self):
        print "pausing Tk processing..."
        self.__pauseRequested.set()

        # 待たないと Tk処理スレッド側でのウィンドウを隠す処理の実施が
        # 確定的にならないので待つ
        self.__iconifyCompleted.acquire() 
        self.__iconifyCompleted.wait()
        self.__iconifyCompleted.release()

    def destroy(self):
        print "waiting for Tk thread to stop..."
        self.__destroyRequested.set()

        # Tk処理スレッドが pause 中であれば
        # self.__startRequested.wait() で待っているので
        self.__startRequested.acquire()
        self.__startRequested.notify()
        self.__startRequested.release()
        
        self.__tkloopThread.join()
        print "finished"

        
    # def waitForQuit(self):
    #     print "waiting for Tk thread to stop..."
    #     if self.__tkloopThread.is_alive():
    #         print "join()..."
    #         self.__tkloopThread.join()
    #     print "finished"
        
        
# テスト用 main ルーチン        
if __name__ == "__main__":
    import time
    
    app = JoyVelCtrlWidget("NOW TESTING...", 0.0, 1.0, 0.0, 1.0)
    time.sleep(2)
    for _ in range(2):
        app.start()
        print "\ncounting down to pause Tk processing..."
        for i in range(3, 0, -1):
            print i
            time.sleep(1)
        app.pause()
        time.sleep(2)
        
    app.destroy()
        
    #app.waitForQuit()



        # if not app.windowClosed():
        #     #print "running"
        #     time.sleep(100)
        # else:
            
    
