﻿#include <cmath>
#include <algorithm>
#include "TRobotUtil.hpp"


static const unsigned char TR_START = 0x02;
static const unsigned char TR_END   = 0x03;
static const unsigned char TR_ROT_VEL_CMD_ID = 0x02;
static const unsigned char TR_RESERVED = 0x00;
static const unsigned char TR_ROT_VEL_CMD_LEN = (unsigned char)TRobot::ROT_VEL_CMD_LEN;
static const unsigned char TR_PREV_DATA_RECV_SUCCESS = 0x99u;
static const unsigned char TR_PREV_DATA_RECV_FAILURE = 0x66u;
static const unsigned char TR_DATA_CHUNK_LEN = (unsigned char)TRobot::BASE_SENSOR_DATA_LEN; 


////////////////////////////////////////////////////////////////////////////////
/*!
 *  ファームウェア用指令シーケンスを組み立てる
 *
 *  @param leftRotVel  左車輪の回転速度指令値（値域: [ROT_VEL_CMD_MIN, ROT_VEL_CMD_MAX]）
 *  @param rightRotVel 右車輪の回転速度指令値（値域: [ROT_VEL_CMD_MIN, ROT_VEL_CMD_MAX]）
 *  @param cmd [出力] 
 *
 *  @return 
 */
TRobot::RetCode TRobot::buildCmdSeqSetRotVel(int leftRotVel, int rightRotVel, unsigned char *cmd)
{
  static int count = 0;
  const int COUNT_MAX = 256;
  
  if((leftRotVel < ROT_VEL_CMD_MIN || ROT_VEL_CMD_MAX < leftRotVel) ||
     (rightRotVel < ROT_VEL_CMD_MIN || ROT_VEL_CMD_MAX < rightRotVel)){
    return TRobot::ERROR_OUT_OF_RANGE;
  }
  cmd[0] = TR_START;
  cmd[1] = TR_ROT_VEL_CMD_ID;
  cmd[2] = TR_ROT_VEL_CMD_LEN;
  cmd[3] = TR_RESERVED;
#if 0 
  cmd[4] = (unsigned char)(leftRotVel & 0xff);
  cmd[5] = (unsigned char)(rightRotVel & 0xff);
#else
  cmd[4] = (unsigned char)(rightRotVel & 0xff);
  cmd[5] = (unsigned char)(leftRotVel & 0xff);
#endif    
  cmd[6] = (unsigned char)(count); 
  cmd[7] = TR_END;

  count = (count + 1) % COUNT_MAX;
  return TRobot::SUCCESS;
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *
 */
bool TRobot::isValidHeader(const unsigned char * const header)
{
  return
    (header[0] == TR_START) &&
    (header[1] == TR_PREV_DATA_RECV_SUCCESS || header[1] == TR_PREV_DATA_RECV_FAILURE) &&
    (header[2] == TR_DATA_CHUNK_LEN);
}


////////////////////////////////////////////////////////////////////////////////
/**
 *
 *  @return 指定されたバッファ中のヘッダの先頭へのポインタ。
 *          ヘッダが見つからなければバッファの末尾(有効部分の終端+1)へのポインタ
 */
const unsigned char *TRobot::searchForValidHeader(const unsigned char *bufFirst,
                                                  const unsigned char *bufLast)
{
  using namespace std;
  const unsigned char *headPos = NULL;

  static const unsigned char HTS[] =
    {TR_START, TR_PREV_DATA_RECV_SUCCESS, TR_DATA_CHUNK_LEN};
  static const unsigned char *HTS_LAST = HTS + sizeof(HTS);
    
  static const unsigned char HTF[] =
    {TR_START, TR_PREV_DATA_RECV_FAILURE, TR_DATA_CHUNK_LEN};
  static const unsigned char *HTF_LAST = HTF + sizeof(HTF);
    
  headPos = search(bufFirst, bufLast, HTS, HTS_LAST);
  if(headPos != bufLast){
    return headPos;
  }
  headPos = search(bufFirst, bufLast, HTF, HTF_LAST);
  if(headPos != bufLast){
    return headPos;
  }
  return bufLast;
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *
 *
 */
TRobot::RetCode TRobot::decodeBaseSensorData(const unsigned char * const buf,
                                             BaseSensorData& sensorData)
{
  const size_t R_CTL_LSB_OFFSET  = 0x03;
  const size_t R_ENC_LSB_OFFSET  = 0x06; 
  const size_t L_ENC_LSB_OFFSET  = 0x07;
  const size_t RL_ENC_MSB_OFFSET = 0x08;
  const size_t CYCLE_CNT_OFFSET  = 0x09;
  const unsigned char R_ENC_MSB_MASK = 0x0f; 
  const unsigned char L_ENC_MSB_MASK = 0xf0; 
  const unsigned char R_ENC_MSB_SIGN_BIT = 0x08; 
  const unsigned char L_ENC_MSB_SIGN_BIT = 0x80; 
  const int U12MAX = 0x0fff; // 12 bit 符号なし整数の最大値
  const size_t R_ENC_MSB_SHIFT = 8;  
  const size_t L_ENC_MSB_SHIFT = 4;
    
  // モータ制御およびエンコーダデータのサンプルの間隔 [bytes]
  const size_t CTL_ENC_SMP_STRIDE = 7;
  const size_t ATT_SEN_SMP_STRIDE = 2; // 姿勢センサの各軸データの間隔 [bytes]
  const size_t ACC_OFFSET = R_CTL_LSB_OFFSET + BaseSensorData::NUM_ENC_COUNTS*CTL_ENC_SMP_STRIDE;
  const size_t GYR_OFFSET = ACC_OFFSET + BaseSensorData::NUM_XYZ*ATT_SEN_SMP_STRIDE;
  const size_t MAG_OFFSET = GYR_OFFSET + BaseSensorData::NUM_XYZ*ATT_SEN_SMP_STRIDE;
  size_t smpIdx = 0;
  unsigned char msb;
  int sgnBits;
  
  for(size_t i = 0; i < BaseSensorData::NUM_ENC_COUNTS; ++i){
    // 符号ビットによって必要に応じて補数表現上妥当なように変換する
    // (符号付き 12bit 整数 -> 符号付き 32bit 整数)
    msb = buf[smpIdx + RL_ENC_MSB_OFFSET];
    
    sgnBits = (msb & R_ENC_MSB_SIGN_BIT)? ~U12MAX : 0x00;
    sensorData.m_rightEnc[i] = sgnBits |
      ((msb & R_ENC_MSB_MASK) << R_ENC_MSB_SHIFT) | buf[smpIdx + R_ENC_LSB_OFFSET];
        
    sgnBits = (msb & L_ENC_MSB_SIGN_BIT)? ~U12MAX : 0x00;
    sensorData.m_leftEnc[i] = sgnBits |
      ((msb & L_ENC_MSB_MASK) << L_ENC_MSB_SHIFT) | buf[smpIdx + L_ENC_LSB_OFFSET];
    sensorData.m_cycleCount[i] = buf[smpIdx + CYCLE_CNT_OFFSET];
    smpIdx += CTL_ENC_SMP_STRIDE;
  }

  // 前提: 姿勢センサのデータは little endian 
  for(size_t ax = 0; ax < BaseSensorData::NUM_XYZ; ++ ax){
    size_t accIdx = ACC_OFFSET + ax*ATT_SEN_SMP_STRIDE;
    size_t gyrIdx = GYR_OFFSET + ax*ATT_SEN_SMP_STRIDE;
    size_t magIdx = MAG_OFFSET + ax*ATT_SEN_SMP_STRIDE;
    sensorData.m_acc[ax] = (int)(buf[accIdx + 1] << 8) | buf[accIdx];
    sensorData.m_gyr[ax] = (int)(buf[gyrIdx + 1] << 8) | buf[gyrIdx];
    sensorData.m_mag[ax] = (int)(buf[magIdx + 1] << 8) | buf[magIdx];
  }

  return TRobot::SUCCESS;
}

